# Import the Portal object.
import geni.portal as portal
# Import the ProtoGENI library.
import geni.rspec.pg as pg
# Import the InstaGENI library.
import geni.rspec.igext as ig
# Import the Emulab specific extensions.
import geni.rspec.emulab as emulab

# Create a portal object,
pc = portal.Context()

pc.defineParameter("vmType","VM Type",portal.ParameterType.STRING,"Docker",
                   [("Docker","Docker"),("Xen","Xen")],
                   longDescription="Choose a VM type (NB: if you force a host image, ensure it is compatible with this choice!)")
pc.defineParameter("numLans","Number LANs",portal.ParameterType.INTEGER,2,
                   longDescription="The number of LANs to create.")
pc.defineParameter("lanSize","Nodes per LAN",portal.ParameterType.INTEGER,8,
                   longDescription="The number of nodes per LAN.")
pc.defineParameter("routerFabric","Routing Fabric",portal.ParameterType.STRING,
                   "mesh",
                   [("mesh","Mesh"),("nMesh","N-way Mesh"),("single","Single"),
                    ("nSpline","N-way Leaf/Spline")],
                   longDescription="The routing fabric connecting the LANs.")
pc.defineParameter("n","Factor",portal.ParameterType.INTEGER,1,
                   longDescription="The factor for the routing fabric type, if any.")
pc.defineParameter("fixNodes","Fix Nodes",portal.ParameterType.BOOLEAN,True,
                   longDescription="Fix each LAN's nodes to a phys host, and all routers to a routerhost.")
pc.defineParameter("shared","Use Shared Containers",portal.ParameterType.BOOLEAN,False,
                   longDescription="Use shared containers if set.  However, if Fix Nodes is set True, this is ignored; if you choose to specify a mapping to dedicated physical machines, your containers will be dedicated, like their physical hosts.")
pc.defineParameter("hostType","Host Node Type",portal.ParameterType.NODETYPE,"",
                   longDescription="Node type of physical host nodes.")
pc.defineParameter("coresPerVM","Cores per VM",portal.ParameterType.INTEGER,0,
                   longDescription="Number of cores each VM will get.")
pc.defineParameter("ramPerVM","RAM per VM",portal.ParameterType.INTEGER,0,
                   longDescription="Amount of RAM each VM will get.")
pc.defineParameter("dockerExternalImages","Images are External Docker",
                   portal.ParameterType.BOOLEAN,False,
                   longDescription='If creating Docker VMs, the specified VM Images are all external Docker images, and a placeholder Cloudlab redirect image will point to them.')
pc.defineParameter("dockerImageFromDockerfiles","Images are Dockerfiles",
                   portal.ParameterType.BOOLEAN,False,
                   longDescription='If creating Docker VMs, the specified images are all Dockerfile URLs, not actual images; we will create and use them on-demand.')
pc.defineParameter("vmImages","VM Images",portal.ParameterType.STRING,
                   'urn:publicid:IDN+emulab.net+image+emulab-ops//UBUNTU16-64-STD',
                   longDescription="A list of images your VMs will run, separated by whitespace.  If more than one, we will alternate through the list as we create VMs.")
pc.defineParameter("storageDriver","Storage Driver",
                   portal.ParameterType.STRING,'',
                   [('',''),('devicemapper','devicemapper'),
                    ('overlay2','overlay2'),('aufs','aufs')],
                   longDescription="The Docker storage driver the host machines will use.")
pc.defineParameter("dockerCmd","Per-node Docker Command",
                   portal.ParameterType.STRING,"",
                   longDescription="As specified in the Docker documentation, this value will either be concatenated with the value of entrypoint in the docker image file; or it will replace it.  Note: we emulate entrypoint/command as a boot service, because we run a real init as the entrypoint.")
pc.defineParameter("dockerEnv","Per-node Docker Environment",
                   portal.ParameterType.STRING,"",
                   longDescription="These environment variables will be placed in a shell script and source by the entrypoint/command emulation.  Shells accept whitespace-separated variable assigment statements; you could also separate them by semicolons if you prefer.  Either way, this value will be placed into a one-line file that will be sourced by another shell script.")
pc.defineParameter("dockerTbAugmentation","Docker TbAugmentation Level",
                   portal.ParameterType.STRING,'default',
                   [('default','Default'),('none','None'),('basic','Basic'),
                    ('core','Core'),('buildenv','BuildEnv'),('full','Full')],
                   longDescription="The requested tbAugmentation level; may be either full, buildenv, core, basic, none.  To augment a Docker image is to take the image and install some or all of the Emulab clientside and dependencies, and other generally useful networking packages, before starting up the nodes that use it, so that it works seamlessly in testbeds based on Emulab.")
pc.defineParameter("dockerTbAugmentationUpdate","Update Docker TbAugmentation",
                   portal.ParameterType.BOOLEAN,False,
                   longDescription="If the image is already augmented, re-augment with the current source tree.")
pc.defineParameter("dockerSshStyle","Docker SSH Style",
                   portal.ParameterType.STRING,'default',
                   [('default','Default'),('direct','Direct'),('exec','Exec')],
                   longDescription="Specify what happens when you ssh to your node; may be direct or exec.  If your container is augmented > basic, and you do not specify this, it defaults to direct.  If your container is not augmented to that level and you do not specify this, it defaults to exec.  direct means that the container is running an sshd inside, and an incoming ssh connection will be handled by the container.  exec means that when you connection, sshd will exec a shell inside your container.  You can change that shell by specifying the docker_exec_shell value.")
pc.defineParameter("dockerExecShell","Docker Exec Shell",
                   portal.ParameterType.STRING,'',
                   longDescription="The shell to run if your Docker ssh_style is direct; otherwise ignored")
pc.defineParameter("startupCmd","VM Startup Command",portal.ParameterType.STRING,"",
                   longDescription="A command each VM will run on startup.")
pc.defineParameter("hostImage","Host Image",portal.ParameterType.STRING,'',
                   longDescription="The image your VM host nodes will run (ensure compat with VM Type choice!)")
pc.defineParameter("clientsideUpdate","Update Clientside",
                   portal.ParameterType.STRING,'no',
                   [('no','No'),('yes','Yes'),('each','Each Boot')],
                   longDescription="If set to Yes, the clientside will be updated on first boot for each vhost prior to booting vnodes.  If set to Each Boot, the clientside will be updated for each boot for each vhost.")
pc.defineParameter("clientsideUpdateRepo","Clientside Update Source Repo",
                   portal.ParameterType.STRING,"",
                   longDescription="If set, the updater will clone this git repo; default is https://gitlab.flux.utah.edu/emulab/emulab-devel.git")
pc.defineParameter("clientsideUpdateRef","Clientside Update Source Ref",
                   portal.ParameterType.STRING,"",
                   longDescription="If set, the updater will checkout this branch or tag instead of master; note that it must be a branch or tag, so that we can see its latest commit hash via ls-remote.")
pc.defineParameter("linkSpeed", "LAN/Link Speed",portal.ParameterType.INTEGER, 1000000,
                   [(0,"Any"),(100000,"100Mb/s"),(1000000,"1Gb/s"),(10000000,"10Gb/s")],
                   longDescription="A specific link speed to use for each LAN and link.")
pc.defineParameter("linkLatency", "LAN/Link Latency",portal.ParameterType.LATENCY, 0,
                   longDescription="A specific latency to use for each LAN and link.")
#pc.defineParameter("linkLoss", "LAN/Link Loss",portal.ParameterType.FLOAT, 0.0,
#                   longDescription="A specific loss rate to use for each LAN and link.")
pc.defineParameter("multiplex", "Multiplex Networks",portal.ParameterType.BOOLEAN, True,
                   longDescription="Multiplex all LANs and links.")
pc.defineParameter("bestEffort", "Best Effort Link Bandwidth",portal.ParameterType.BOOLEAN, True,
                   longDescription="Do not require guaranteed bandwidth throughout switch fabric.")
pc.defineParameter("trivialOk", "Trivial Links Ok",portal.ParameterType.BOOLEAN, True,
                   longDescription="Maybe use trivial links.")
#pc.defineParameter("forceVlanLinkType", "Force VLAN Link Type",portal.ParameterType.BOOLEAN, False,
#                   longDescription="Force the type of each LAN/Link to vlan.")
pc.defineParameter("ipAllocationStrategy","IP Addressing",
                   portal.ParameterType.STRING,"script",[("cloudlab","CloudLab"),("script","This Script")],
                   longDescription="Either let CloudLab auto-generate IP addresses for the nodes in your networks, or let this script generate them.  If you include nodes at multiple sites, you must choose this script!  The default is this script, because the subnets CloudLab generates for flat networks are sized according to the number of physical nodes in your topology.  If the script IP address generation is buggy or otherwise insufficient, you can fall back to CloudLab and see if that improves things.")
pc.defineParameter("ipAllocationUsesCIDR","Script IP Addressing Uses CIDR",
                   portal.ParameterType.BOOLEAN,False,
                   longDescription="If this script is generating the IP addresses for your nodes, and this button is checked, we will generate a minimal set of subnets that are just large enough to contain the hosts in each subnet.  If it is not checked, we will generate class B/C subnets inside the 10.0.0.0/8 class A unrouted subnet, for each separate link/lan.")

params = pc.bindParameters()

if params.coresPerVM < 0:
    pc.reportError(portal.ParameterError("Must specify at least one core per VM",['coresPerVM']))
if params.ramPerVM < 0:
    pc.reportError(portal.ParameterError("Must specify at least one core per VM",['ramPerVM']))


# Create a Request object to start building the RSpec.
request = pc.makeRequestRSpec()

tour = ig.Tour()
tour.Description(ig.Tour.TEXT,"Instantiate one or many dedicated VMs with many potential parameters and complex topologies.")
request.addTour(tour)

lans = list()
for i in range(0,params.numLans):
    lan = pg.LAN('lan-%d' % (i,))
    if params.trivialOk is True:
        lan.trivial_ok = True
    else:
        lan.trivial_ok = False
    if params.bestEffort is True:
        lan.best_effort = True
    else:
        lan.best_effort = False
    if params.multiplex is True:
        lan.link_multiplexing = True
    else:
        lan.link_multiplexing = False
    #if params.forceVlanLinkType:
    #    lan.type = "vlan"
    lan.vlan_tagging = True
    if params.linkSpeed > 0:
        lan.bandwidth = params.linkSpeed
    if params.linkLatency > 0:
        lan.latency = params.linkLatency
    #if params.linkLoss > 0:
    #    lan.loss = params.linkLoss
    lans.append(lan)
    pass

if not params.hostImage:
    if params.vmType == 'Docker':
        params.hostImage = 'urn:publicid:IDN+emulab.net+image+emulab-ops//DOCKVH-CE17-1-27-STD'
    elif params.vmType == 'Xen':
        params.hostImage = 'urn:publicid:IDN+emulab.net+image+emulab-ops//XEN44-64-STD'
    pass

hosts = list()
routerHost = None
vnodes = dict()
links = list()
imageList = []
if params.vmImages:
    imageList = params.vmImages.split(" ")
    tmplist = []
    for x in imageList:
        if x and params.vmType == 'Docker' and params.dockerExternalImages:
            if x == 'urn:publicid:IDN+emulab.net+image+emulab-ops//UBUNTU16-64-STD':
                x = 'ubuntu:16.04'
            elif x == 'urn:publicid:IDN+emulab.net+image+emulab-ops//CENTOS7-STD':
                x = 'centos:7'
        if x:
            tmplist.append(x)
        pass
    imageList = tmplist
    pass

#
# Build the vhosts and routerhost, if any.
#
hostNames = []
if params.fixNodes:
    for i in range(0,params.numLans):
        hostNames.append('vhost-%d' % (i,))
    hostNames.append('routerhost')
    for name in hostNames:
        host = pg.RawPC(name)
        host.exclusive = True
        if params.hostType:
            host.hardware_type = params.hostType
        if params.hostImage:
            host.disk_image = params.hostImage
            pass
        if params.clientsideUpdate == "yes" or params.clientsideUpdate == "each":
            host.Attribute("CLIENTSIDE_UPDATE","1")
            if params.clientsideUpdate == "each":
                host.Attribute("CLIENTSIDE_UPDATE_EACHBOOT","1")
            if params.clientsideUpdateRepo:
                host.Attribute("CLIENTSIDE_UPDATE_REPO",params.clientsideUpdateRepo)
            if params.clientsideUpdateRef:
                host.Attribute("CLIENTSIDE_UPDATE_REF",params.clientsideUpdateRef)
        if params.storageDriver:
            host.Attribute("DOCKER_STORAGE_DRIVER",params.storageDriver)
            pass
        hosts.append(host)
    pass

#
# Build the vnodes.
#
for i in range(0,params.numLans):
    for j in range(0,params.lanSize):
        name = 'node-%d-%d' % (i,j)
        vhostname = 'vhost-%d' % (i,)
        if params.vmType == 'Docker':
            node = ig.DockerContainer(name)
        else:
            node = ig.XenVM(name)
        if params.coresPerVM > 0:
            node.cores = params.coresPerVM
        if params.ramPerVM:
            node.ram = params.ramPerVM
        if params.hostType and params.vmType == 'Xen':
            node.xen_ptype = '%s-vm' % (params.hostType,)
        elif params.hostType and params.vmType == 'Docker':
            node.docker_ptype = '%s-vm' % (params.hostType,)
        if params.fixNodes:
            node.InstantiateOn(vhostname)
            node.exclusive = True
        else:
            node.exclusive = not params.shared
        if params.startupCmd:
            node.addService(pg.Execute(shell="sh",command=params.startupCmd))
        if params.vmType == 'Xen' and len(imageList) > 0:
            node.disk_image = imageList[j % len(imageList)]
        elif params.vmType == 'Docker':
            if params.dockerExternalImages == True:
                node.disk_image = 'urn:publicid:IDN+emulab.net+image+emulab-ops//DOCKER-EXT'
                if len(imageList) > 0:
                    node.docker_extimage = imageList[j % len(imageList)]
            elif params.dockerImageFromDockerfiles == True:
                node.disk_image = 'urn:publicid:IDN+emulab.net+image+emulab-ops//DOCKER-EXT'
                if len(imageList) > 0:
                    node.docker_dockerfile = imageList[j % len(imageList)]
            elif len(imageList) > 0:
                node.disk_image = imageList[j % len(imageList)]
            if params.dockerTbAugmentation != 'default':
                node.docker_tbaugmentation = params.dockerTbAugmentation
            if params.dockerTbAugmentationUpdate:
                node.docker_tbaugmentation_update = params.dockerTbAugmentationUpdate
            if params.dockerSshStyle != 'default':
                node.docker_ssh_style = params.dockerSshStyle
            if params.dockerExecShell:
                node.docker_exec_shell = params.dockerExecShell
            if params.dockerCmd:
                node.docker_cmd = params.dockerCmd
                #node.Attribute("DOCKER_CMD",params.dockerCmd)
            if params.dockerEnv:
                node.docker_env = params.dockerEnv
                #node.Attribute("DOCKER_ENV",params.dockerEnv)
            pass

        iface = node.addInterface('if-%d-%d' % (i,j))
        lans[i].addInterface(iface)

        if not vhostname in vnodes:
            vnodes[vhostname] = []
        vnodes[vhostname].append(node)
    pass

def setLinkParams(link):
    if params.trivialOk:
        link.trivial_ok = True
    else:
        link.trivial_ok = False
    if params.bestEffort:
        link.best_effort = True
    if params.multiplex:
        link.link_multiplexing = True
    link.vlan_tagging = True
    #if params.forceVlanLinkType:
    #    link.type = "vlan"

    if params.linkSpeed > 0:
        link.bandwidth = params.linkSpeed
    if params.linkLatency > 0:
        link.latency = params.linkLatency
    #if params.linkLoss > 0:
    #    link.loss = params.linkLoss
    return

#
# Build the routers and fabric.
#
routers = list()
links = list()
nRouters = params.numLans
if params.routerFabric == "single":
    nRouters = 1
for i in range(0,nRouters):
    name = 'router-%d' % (i,)
    if params.vmType == 'Docker':
        node = ig.DockerContainer(name)
    else:
        node = ig.XenVM(name)
    if params.coresPerVM > 0:
        node.cores = params.coresPerVM
    if params.ramPerVM:
        node.ram = params.ramPerVM
    if params.hostType and params.vmType == 'Xen':
        node.xen_ptype = '%s-vm' % (params.hostType,)
    elif params.hostType and params.vmType == 'Docker':
        node.docker_ptype = '%s-vm' % (params.hostType,)
    if params.fixNodes:
        node.InstantiateOn('routerhost')
        node.exclusive = True
    else:
        node.exclusive = not params.shared
    if params.startupCmd:
        node.addService(pg.Execute(shell="sh",command=params.startupCmd))
    if params.vmType == 'Xen' and len(imageList) > 0:
        node.disk_image = imageList[j % len(imageList)]
    elif params.vmType == 'Docker':
        if params.dockerExternalImages == True:
            node.disk_image = 'urn:publicid:IDN+emulab.net+image+emulab-ops//DOCKER-EXT'
            if len(imageList) > 0:
                node.docker_extimage = imageList[j % len(imageList)]
        elif params.dockerImageFromDockerfiles == True:
            node.disk_image = 'urn:publicid:IDN+emulab.net+image+emulab-ops//DOCKER-EXT'
            if len(imageList) > 0:
                node.docker_dockerfile = imageList[j % len(imageList)]
        elif len(imageList) > 0:
            node.disk_image = imageList[j % len(imageList)]
        if params.dockerTbAugmentation != 'default':
            node.docker_tbaugmentation = params.dockerTbAugmentation
        if params.dockerTbAugmentationUpdate:
            node.docker_tbaugmentation_update = params.dockerTbAugmentationUpdate
        if params.dockerSshStyle != 'default':
            node.docker_ssh_style = params.dockerSshStyle
        if params.dockerExecShell:
            node.docker_exec_shell = params.dockerExecShell
        if params.dockerCmd:
            node.docker_cmd = params.dockerCmd
            #node.Attribute("DOCKER_CMD",params.dockerCmd)
        if params.dockerEnv:
            node.docker_env = params.dockerEnv
            #node.Attribute("DOCKER_ENV",params.dockerEnv)
        pass

    routers.append(node)
    pass

#
# Connect the router(s) to their LANs
#
if params.routerFabric == "single":
    for i in range(0,params.numLans):
        niface = vnodes['vhost-%d' % (i,)][0].addInterface('ifR')
        riface = routers[0].addInterface('ifL' % (k,))
        rlink = pg.Link("rlink-%d" % (i,))
        rlink.addInterface(niface)
        rlink.addInterface(riface)
        setLinkParams(rlink)
        links.append(rlink)
else:
    for i in range(0,params.numLans):
        niface = vnodes['vhost-%d' % (i,)][0].addInterface('ifR-%d' % (i,))
        riface = routers[i].addInterface('ifL-%d' % (i,))
        rlink = pg.Link("rlink-%d" % (i,))
        rlink.addInterface(niface)
        rlink.addInterface(riface)
        setLinkParams(rlink)
        links.append(rlink)
    pass

#
# Connect the routers to each other.
#
if params.routerFabric == "mesh":
    meshlinksdone = dict()
    for i in range(0,params.numLans):
        meshlinksdone[i] = dict()
    for i in range(0,params.numLans):
        for j in range(0,params.numLans):
            if j >= i:
                continue
            meshlinksdone[i][j] = False
    for i in range(0,params.numLans):
        irouter = routers[i]
        for j in range(0,params.numLans):
            if i == j or j >= i:
                continue
            jrouter = routers[j]
            iiface = irouter.addInterface("ifR-%d-%d" % (i,j))
            jiface = jrouter.addInterface("ifR-%d-%d" % (j,i))
            mlink = pg.Link("mlink-%d-%d" % (i,j))
            mlink.addInterface(iiface)
            mlink.addInterface(jiface)
            setLinkParams(mlink)
            links.append(mlink)
    pass
elif params.routerFabric == "single":
    pass
else:
    pc.reportError(portal.ParameterError("Invalid value for Routing Fabric!",
                                         ['routerFabric']))

#
# Addressing utils.
#
baseaddr = [10,0,0,0]
basebits = 8
minsubnetbits = basebits + 1
maxsubnetbits = 32 - 2
networks = dict()
netmasks = dict()
lastaddr = dict()
# The list of allowed subnet bit sizes.
ALLOWED_SUBNETS = []
# An ordered (increasing bit size) list of (bitsize,networkname) tuples.
REQUIRED_SUBNETS = []
# A dict of key networkname and value { 'bits':bitsize,'network':networkint,
#   'hostnum':hostint,'maxhosts':maxhostsint,'netmaskstr':'x.x.x.x' }
SUBNETS = {}

if params.ipAllocationUsesCIDR:
    subbits = basebits + 1
    while subbits <= maxsubnetbits:
        ALLOWED_SUBNETS.append(subbits)
        subbits += 1
else:
    ALLOWED_SUBNETS = [ 16,24 ]

def __htoa(h):
    return "%d.%d.%d.%d" % (
        (h >> 24) & 0xff,(h >> 16) & 0xff,(h >> 8) & 0xff,h & 0xff)

def ipassign_request_network(name,numhosts):
    #
    # Here, we just reserve each network's bitspace, in increasing
    # bitsize.  This makes eventual assignment trivial; we just pick the
    # next hole out of the previous (larger or equiv size) subnet.
    #
    _sz = numhosts + 1
    i = 0
    while _sz > 0:
        i += 1
        _sz = _sz >> 1
    # The minimum network size is a 255.255.255.252 (a /30).
    if i < 2:
        i = 2
    # Change the host bitsize to a network bitsize.
    i = 32 - i
    sn = None
    for _sn in ALLOWED_SUBNETS:
        if i < _sn:
            break
        sn = _sn
    if sn == None:
        raise Exception("network of bitsize %d (%d hosts) cannot be supported"
                        " with basebits %d" % (i,numhosts,basebits))
    ipos = 0
    for (bitsize,_name) in REQUIRED_SUBNETS:
        if sn < bitsize:
            break
        ipos += 1
    REQUIRED_SUBNETS.insert(ipos,(sn,name))
    return

def ipassign_assign_networks():
    basenum = baseaddr[0] << 24 | baseaddr[1] << 16 \
        | baseaddr[2] << 8 | baseaddr[3]
    # If any of our network broadcast addrs exceed this address, we've
    # overallocated.
    nextbasenum = basenum + 2 ** (32 - basebits)

    #
    # Process REQUIRED_SUBNETS list into SUBNETS dict, and setup the
    # next host number.  Again, we assume that REQUIRED_SUBNETS is
    # ordered in increasing number of network bits required.
    #
    lastbitsize = None
    lastnetnum = None
    for i in range(0,len(REQUIRED_SUBNETS)):
        (bitsize,name) = REQUIRED_SUBNETS[i]
        if lastbitsize != bitsize and lastbitsize != None:
            # Need to get into the next lastbitsize subnet to allocate
            # for the bitsize subnet.
            lastnetnum += 2 ** (32 - lastbitsize)

        if lastnetnum != None:
            netnum = lastnetnum + 2 ** (32 - bitsize)
        else:
            netnum = basenum
        if netnum >= nextbasenum:
            raise Exception("out of network space in /%d at network %s (/%d)"
                            % (basebits,name,bitsize))
        networknum = netnum + (2 ** (32 - bitsize)) - 1
        netmask = 0
        for i in range(32 - bitsize,32):
            netmask |= (1 << i)
        SUBNETS[name] = {
            'bits':bitsize,'netnum':netnum,
            'networknum':networknum,'networkstr':__htoa(networknum),
            'hostnum':0,'maxhosts':2 ** (32 - bitsize) - 1,'addrs':[],
            'netmask':netmask,'netmaskstr':__htoa(netmask)
        }
        lastnetnum = netnum
        lastbitsize = bitsize

def ipassign_assign_host(lan):
    #
    # Assign an IP address in the given lan.  If you call this function
    # more times than you effectively promised (by calling
    # ipassign_add_network with a requested host size), you will be
    # handed an Exception.
    #
    bitsize = None
    for (_bitsize,_name) in REQUIRED_SUBNETS:
        if _name == lan:
            bitsize = _bitsize
    if not bitsize:
        raise Exception("unknown network name %s" % (lan,))
    if not _name in SUBNETS:
        raise Exception("unknown network name %s" % (lan,))

    maxhosts = (2 ** bitsize - 1)
    if SUBNETS[lan]['hostnum'] >= maxhosts:
        raise Exception("no host space left in network name %s (max hosts %d)"
                        % (lan,maxhosts))

    SUBNETS[lan]['hostnum'] += 1

    addr = __htoa(SUBNETS[lan]['netnum'] + SUBNETS[lan]['hostnum'])
    if True:
        SUBNETS[lan]['addrs'].append(addr)
    return (addr,SUBNETS[lan]['netmaskstr'])

#
# Assign IP addresses if necessary.
#
if params.ipAllocationStrategy == 'script':
    for lan in lans:
        name = lan.client_id
        ipassign_request_network(name,len(lan.interfaces))
    for link in links:
        name = link.client_id
        ipassign_request_network(name,len(link.interfaces))
    ipassign_assign_networks()
    for lan in lans:
        name = lan.client_id
        for iface in lan.interfaces:
            (address,netmask) = ipassign_assign_host(name)
            iface.addAddress(pg.IPv4Address(address,netmask))
    for link in links:
        name = link.client_id
        for iface in link.interfaces:
            (address,netmask) = ipassign_assign_host(name)
            iface.addAddress(pg.IPv4Address(address,netmask))
    pass

#
# If this is a LAN-only experiment, turn off ddjikstra routing; we don't
# need it, and it explodes on huge LANs (i.e., on a 5k topo, 30s per vnode).
#
#if not params.buildSnakeLink:
#    request.setRoutingStyle("none")

for x in hosts:
    request.addResource(x)
for x in vnodes:
    for y in vnodes[x]:
        request.addResource(y)
for x in routers:
    request.addResource(x)
for x in lans:
    request.addResource(x)
for x in links:
    request.addResource(x)

pc.printRequestRSpec(request)
